from setuptools import setup, find_packages

setup(
    name="mctech_discovery",
    version="1.2.0",
    packages=find_packages(
        include=["mctech_discovery.*"],
        exclude=["*.tests"]
    ),
    exclude_package_data={"mctech_discovery": ["testmain.py"]},
    install_requires=["log4py", "netifaces",
                      "py_eureka_client", "httpx",
                      "mctech-actuator", "mctech-core", "websocket-client"]
)



from __future__ import absolute_import
import websocket
from typing import Iterable, AsyncIterable
from abc import ABC, abstractmethod

from ..request_context import RequestContext
from ..ep import EndPointRequest

from .._rpc_types import Invoker, RpcServiceInfo, EndPoint, MixedMethod, RpcResponseResult, CallMode


class LoadBalancer(ABC):
    @abstractmethod
    def create_context(self,
                       method: MixedMethod,
                       invoker: Invoker,
                       service: RpcServiceInfo) -> RequestContext:
        pass

    @abstractmethod
    def check_alive(self) -> bool:
        pass

    def request_async(self, method: CallMode, req: EndPointRequest):
        if method == 'execute':
            return self.execute_async(req)
        elif method == 'stream':
            return self.stream_async(req)
        elif method == 'ws':
            return self.ws_async(req)  # 底层创建websocket用的是同步方法，当前返回值会转换成异步调用的包装
        raise RuntimeError('不支持的方法:' + method)

    def request(self, method: CallMode, req: EndPointRequest):
        if method == 'execute':
            return self.execute(req)
        elif method == 'stream':
            return self.stream(req)
        elif method == 'ws':
            return self.ws(req)  # 底层创建websocket用的是同步方法，当前返回值会转换成异步调用的包装
        raise RuntimeError('不支持的方法:' + method)

    @abstractmethod
    async def execute_async(self, req: EndPointRequest) -> RpcResponseResult:
        pass

    @abstractmethod
    async def stream_async(self, req: EndPointRequest) -> AsyncIterable[bytes]:
        pass

    @abstractmethod
    async def ws_async(self, req: EndPointRequest) -> websocket.WebSocket:
        pass

    @abstractmethod
    def execute(self, req: EndPointRequest) -> RpcResponseResult:
        pass

    @abstractmethod
    def stream(self, req: EndPointRequest) -> Iterable[bytes]:
        pass

    @abstractmethod
    def ws(self, req: EndPointRequest) -> websocket.WebSocket:
        pass


class ServiceLoadBalancer(LoadBalancer):
    @abstractmethod
    def choose_one(self) -> EndPoint:
        pass



from __future__ import absolute_import
import websocket
from typing import Iterable, AsyncIterable

from .ep_request import EndPointRequest
from .internal_rpc import execute as rpc_execute, stream as rpc_stream, ws as rpc_ws, \
    execute_async as rpc_execute_async, stream_async as rpc_stream_async, ws_async as rpc_ws_async

from .._rpc_types import EndPoint, RpcResponseResult


class SimpleEndPointRequest (EndPointRequest):
    async def execute_async(self, ep: EndPoint) -> RpcResponseResult:
        resultData = await rpc_execute_async(self.context, ep)
        return resultData

    async def stream_async(self, ep: EndPoint) -> AsyncIterable[bytes]:
        resultStream = await rpc_stream_async(self.context, ep)
        return resultStream

    async def ws_async(self, ep: EndPoint) -> websocket.WebSocket:
        conn = await rpc_ws_async(self.context, ep)
        return conn

    def execute(self, ep: EndPoint) -> RpcResponseResult:
        resultData = rpc_execute(self.context, ep)
        return resultData

    def stream(self, ep: EndPoint) -> Iterable[bytes]:
        resultStream = rpc_stream(self.context, ep)
        return resultStream

    def ws(self, ep: EndPoint) -> websocket.WebSocket:
        conn = rpc_ws(self.context, ep)
        return conn



from __future__ import absolute_import
import websocket
from typing import Iterable, AsyncIterable

from ._lb_types import LoadBalancer
from ..request_context import RequestContext
from ..ep import EndPointRequest

from .._rpc_types import Invoker, RpcServiceInfo, MixedMethod, RpcResponseResult


class RetryLoadBalancer(LoadBalancer):

    def __init__(self, lb: LoadBalancer):
        super().__init__()
        self._lb = lb

    def check_alive(self) -> bool:
        return self._lb.check_alive()

    def create_context(self,
                       method: MixedMethod,
                       invoker: Invoker,
                       service: RpcServiceInfo) -> RequestContext:
        return self._lb.create_context(method, invoker, service)

    async def execute_async(self, req: EndPointRequest) -> RpcResponseResult:
        result_data = await self._lb.execute_async(req)
        return result_data

    async def stream_async(self, req: EndPointRequest) -> AsyncIterable[bytes]:
        result_stream = await self._lb.stream_async(req)
        return result_stream

    async def ws_async(self, req: EndPointRequest) -> websocket.WebSocket:
        conn = await self._lb.ws_async(req)
        return conn

    def execute(self, req: EndPointRequest) -> RpcResponseResult:
        result_data = self._lb.execute(req)
        return result_data

    def stream(self, req: EndPointRequest) -> Iterable[bytes]:
        result_stream = self._lb.stream(req)
        return result_stream

    def ws(self, req: EndPointRequest) -> websocket.WebSocket:
        conn = self._lb.ws(req)
        return conn



from __future__ import absolute_import
import asyncio
import re
from functools import partial
from typing import Optional, Dict, Mapping, Tuple, Any, Iterable, AsyncIterable
from httpx import URL, Request, Response

from .httpx_clients import clients, async_clients
from ..web_error import WebError
from ..request_context import RequestContext
from ...discovery import get_discovery

from .._rpc_types import Invoker, RpcInvoker, EndPoint, HttpRequestBody, RequestOptions, \
    WebSocketRequestOptions, RpcRequestOptions, \
    ResponseTypes, RpcResponseResult

_loop = asyncio.new_event_loop()

_CONTENT_TYPE_VALUE = 'application/json; charset=UTF-8'


def _create_websocket_options(context: RequestContext, ep: EndPoint) -> Tuple[str, Mapping[str, Any]]:
    option = _create_request_options(context, ep, None)

    address = URL(option.url)
    if address.scheme == 'https':
        address = URL(context.url, scheme='wss')
    else:
        address = URL(context.url, scheme='ws')
    headers = option.headers
    import ssl
    ws_options = {
        'header': headers,
        'timeout': option.timeout,
        'sslopt': {"cert_reqs": ssl.CERT_NONE} if option.verify else None,
        'host': headers.get('Host')
    }
    return (str(address), ws_options)


def _create_httpx_request(context: RequestContext, ep: EndPoint, url: Optional[str]) -> Tuple[bool, Request]:
    option = _create_request_options(context, ep, url)
    assert isinstance(option, RpcRequestOptions)

    kwargs = {
        'url': option.url,
        'method': option.method,
        'headers': dict(option.headers),
        'timeout': clients.create_timeout(option.timeout)
    }
    headers: Dict[str, Any] = kwargs['headers']
    if option.body is not None:
        body = option.body
        content_type = headers.get('Content-Type') or ''
        if isinstance(body, HttpRequestBody):
            name, value = body.args()
            if name:
                kwargs[name] = value
            if body.content_type:
                # body上的content_type优先
                headers['Content-Type'] = body.content_type
        elif isinstance(body, (str, bytes)):
            kwargs['content'] = body
        elif content_type.lower().find('json'):
            kwargs['json'] = body
        else:
            raise RuntimeError('不支持的数据类型: %s' % type(body).__name__)
        if 'Content-Type' not in headers:
            # content-type 不存在，设置默认值
            headers['Content-Type'] = _CONTENT_TYPE_VALUE

    return (option.verify, clients.build_request(**kwargs))


def _create_request_options(context: RequestContext, ep: EndPoint, url: Optional[str]) -> RequestOptions:
    target_url = context.url
    if not ep.vip_address:
        # URL构造函数支持 URL参数类型
        target_url = URL(context.url, scheme="http", host=ep.host, port=ep.port)

    kwargs = {
        'headers': _normalize_headers(context.headers),
        'url': url or target_url,
        'verify': context.service.reject_unauthorized,
        'timeout': context.timeout
    }

    if context.method == 'ws':
        return WebSocketRequestOptions(**kwargs)

    invoker = context.invoker
    assert isinstance(invoker, RpcInvoker)

    kwargs['method'] = context.method
    kwargs['body'] = invoker.body
    kwargs['response_type'] = invoker.response_type or ('json' if invoker.json else 'text')
    reqOption = RpcRequestOptions(**kwargs)
    context.process_request_option(reqOption)
    return reqOption


__HEADER_PATTERN = re.compile('(-|^)[a-z]', re.IGNORECASE)


def _normalize_headers(headers: Mapping[str, Any]) -> Mapping[str, Any]:
    new_headers = {}
    if not headers:
        return new_headers

    for key, value in headers.items():
        new_key = __HEADER_PATTERN.sub(lambda m: m.group(0).upper(), key)
        new_headers[new_key] = value
    return new_headers


def resolve_error(context: RequestContext, res: Response):
    timeout = context.timeout if context.timeout else 0
    service = context.service

    message = "%s --> %s --> [timeout:%ds, cost: %fs] [%s] [%s] %s" % (
        res.reason_phrase, res.text, timeout, res.elapsed.total_seconds(),  # noqa
        service.service_id, context.method, context.url)
    err = WebError(message, status=res.status_code)
    context.resolve_error(res, err)
    raise err


async def _walk_using_urllib_async(context: RequestContext, ep: EndPoint, stream: bool, url: Optional[str] = None) -> Response:  # noqa
    verify, req = _create_httpx_request(context, ep, url)
    res = await async_clients.get_client(verify).send(req, stream=stream)
    return res


def _walk_using_urllib(context: RequestContext, ep: EndPoint, stream: bool, url: Optional[str] = None) -> Response:  # noqa
    verify, req = _create_httpx_request(context, ep, url)
    res = clients.get_client(verify).send(req, stream=stream)
    return res


async def _do_request_async(context: RequestContext, ep: EndPoint, stream: bool) -> Response:
    if not ep.vip_address:
        url = ep.to_url(path=context.url.path, params=context.url.params)
        res = await _walk_using_urllib_async(context, ep, stream, str(url))
    else:
        res = await get_discovery().client.walk_nodes(
            app_name=context.service.service_id,
            service=context.url.path,
            walker=partial(_walk_using_urllib_async, context, ep, stream)
        )

    if res.is_error:
        try:
            if res.stream and not res.is_stream_consumed:
                # stream方式且未读取过，先读取到content里
                await res.aread()
            resolve_error(context, res)
        finally:
            # 处理完异常后可关闭
            await res.aclose()
    elif not res.stream:
        # 非stream模式，已经读取完所有内容，可关闭
        res.aclose()
    return res


def _do_request(context: RequestContext, ep: EndPoint, stream: bool) -> Response:
    if ep:
        url = ep.to_url(path=context.url.path, params=context.url.params)
        res = _walk_using_urllib(context, ep, stream, str(url))
    else:
        r = get_discovery().client.walk_nodes(
            app_name=context.service.service_id,
            service=context.path_and_query,
            walker=partial(_walk_using_urllib, context, ep, stream)
        )
        res = _loop.run_until_complete(r)

    if res.is_error:
        try:
            if res.stream and not res.is_stream_consumed:
                # stream方式且未读取过，先读取到content里
                res.read()
            resolve_error(context, res)
        finally:
            # 处理完异常后可关闭
            res.close()
    elif not res.stream:
        # 非stream模式，已经读取完所有内容，可关闭
        res.close()
    return res

#####################################################################################


def _convert_result(res: Response, invoker: Invoker) -> RpcResponseResult:
    assert isinstance(invoker, RpcInvoker)
    response_type: ResponseTypes = 'json' if invoker.json else 'text'
    if invoker.response_type:
        response_type = invoker.response_type

    content_type = res.headers.get('content-type') or ''
    if response_type == 'json':
        if content_type.find('json') >= 0:
            return res.json()
        else:
            return res.text
    elif response_type == 'text':
        return res.text
    elif response_type == 'bytes':
        return res.read()
    return res


def execute(context: RequestContext, ep: EndPoint) -> RpcResponseResult:
    res = _do_request(context, ep, False)
    return _convert_result(res, context.invoker)


def stream(context: RequestContext, ep: EndPoint) -> Iterable[bytes]:
    res = _do_request(context, ep, True)
    return res.iter_bytes()


def ws(context: RequestContext, ep: EndPoint):
    address, option = _create_websocket_options(context, ep)
    from websocket import create_connection
    conn = create_connection(url=address, **option)

    return conn

###########################################################################################


async def execute_async(context: RequestContext, ep: EndPoint) -> RpcResponseResult:
    res = await _do_request_async(context, ep, False)
    return _convert_result(res, context.invoker)


async def stream_async(context: RequestContext, ep: EndPoint) -> AsyncIterable[bytes]:
    res = await _do_request_async(context, ep, True)
    return res.aiter_bytes()


async def ws_async(context: RequestContext, ep: EndPoint):
    return ws(context, ep)



from __future__ import absolute_import

import re
from abc import ABC, abstractmethod
from typing import Dict, Any, Tuple, Mapping, Optional
from httpx import URL, Response
from urllib.parse import urljoin

from mctech_core.context import get_async_context

from .web_error import WebError

from ._rpc_types import Invoker, RpcInvoker, RpcServiceInfo, Seconds, RpcRequestOptions, MixedMethod

PARAM_PARTERN = re.compile(':(\\w+)')
HEADER_PARTTERN = re.compile('[A-Z]')


class RequestContext(ABC):
    method: MixedMethod
    invoker: Invoker
    service: RpcServiceInfo
    app_name: Optional[str]
    headers: Mapping[str, Any]
    url: URL

    def __init__(self, method: MixedMethod, invoker: Invoker, service: RpcServiceInfo,
                 app_name: Optional[str] = None,
                 product_line: Optional[str] = None):
        self.method = method
        self.invoker = invoker
        self.service = service
        self.app_name = app_name
        self.product_line = product_line

        path = invoker.path
        query = invoker.query if invoker.query is not None else {}
        params = invoker.params if invoker.params is not None else {}

        headers = dict(invoker.headers) if invoker.headers else {}
        if invoker.context:
            self._add_headers(headers, invoker.context)
        self.headers = headers

        base_url, path_and_query = self._get_path_and_query(path)
        url = URL(base_url).join(path_and_query)
        url = url.copy_merge_params(query)

        def _replacement(matched, params):
            var_name: str = matched.group(1)
            # 获取传入参数
            value = params.get(var_name)
            if (value is None):
                error_msg = "不存在的属性：'%s'" % var_name
                raise RuntimeError(error_msg)
            return str(value)

        new_path = PARAM_PARTERN.sub(
            lambda matched: _replacement(matched, params), url.path)
        url = url.copy_with(path=new_path)
        self.url = url

    @property
    def timeout(self) -> Optional[Seconds]:
        if isinstance(self.invoker, RpcInvoker) and self.invoker.timeout:
            return self.invoker.timeout
        return self.service.timeout

    def process_request_option(self, option: RpcRequestOptions):
        # 什么也不做
        pass

    @abstractmethod
    def resolve_error(self, res: Response, err: WebError):
        pass

    @abstractmethod
    def _get_path_and_query(self, path: str) -> Tuple[str, str]:
        '''
        :return [baseUrl: str, path_and_query: str]
        '''
        pass

    @abstractmethod
    def _add_headers(self, headers: Dict[str, Any], context: Mapping[str, Any]):
        pass

    def _to_header_name(self, name: str, prefix: str):
        key = HEADER_PARTTERN.sub(
            lambda matched: '-' + matched.group(0).upper(), name)

        # 补充'x-'前缀
        if key.startswith('-'):
            return prefix + key
        return prefix + '-' + key


class InternalRequestContext(RequestContext):
    def _get_path_and_query(self, path: str):
        base_url = self.service.url
        if not base_url:
            base_url = "http://%s" % self.service.service_id

        path_prefix = self.service.path if self.service.path else '/'
        path_and_query = urljoin(path_prefix, path)
        return (base_url, path_and_query)

    def _add_headers(self, headers: Dict[str, Any], context: Mapping[str, Any]):
        asynctx = get_async_context()
        principal: Mapping[str, Any] = {}
        extras: Mapping[str, Any] = {}
        webContext = asynctx.web_context
        if webContext:
            # 当前调用是在处理web请求的时候发生的
            if webContext.principal:
                # website身份认证信息
                principal['id'] = webContext.principal.get('id')
                if not principal.get('userId') and principal.get('id'):
                    principal['userId'] = principal['id']

                principal['tenantId'] = webContext.principal['tenantId']
                principal['tenantCode'] = webContext.principal['tenantCode']

                # TODO: 应该删除掉
                principal['orgId'] = webContext.principal['orgId']
                principal['applicationId'] = webContext.principal['applicationId']  # noqa
            extras = webContext.extras

        if not principal.get('tenantId'):
            # 当前调用是在消息队列或定时任务等其它非web请求中调用的
            # 消息队列处理器中中没有webContext对象
            principal['tenantId'] = asynctx.tenant_id

        mergedCtx = {**principal, **extras}
        # 用户调用api时显示设置的信息
        if context:
            mergedCtx.update(context)

        for name, value in mergedCtx.items():
            if value is None:
                continue
            key = self._to_header_name(name, 'x')
            headers[key] = str(value)

        if self.app_name:
            headers['i-rpc-client'] = self.app_name

        tracing = asynctx.tracing
        if not tracing:
            return

        # 用于调用链跟踪的信息
        for name, value in asynctx.tracing.items():
            headers[name] = value

    def resolve_error(self, res: Response, err: WebError):
        content_type = res.headers.get('content-type') or ''
        err.code = 'HTTP_CLIENT_ERROR' if res.status_code < 500 else 'HTTP_SERVER_ERROR'
        err.desc = '调用服务发生错误'
        err.status = res.status_code

        if content_type.find('json') >= 0:
            try:
                data = res.json()
            except Exception:
                # 什么也不做
                data: Mapping[str, Any] = {
                    'code': 'json_format_error',
                    'desc': res.text
                }
            # 默认json格式
            err.code = data.get("code")
            err.desc = data.get("desc")
            err.details = data.get("details") or []
        else:
            err.code = 'HTTP_CLIENT_ERROR' if res.status_code < 500 else 'HTTP_SERVER_ERROR'
            err.desc = res.text



from __future__ import absolute_import
import os

from fastapi.responses import JSONResponse
from typing import Dict
from urllib.parse import urlparse
from threading import Timer

from mctech_core import get_configure
from mctech_actuator import get_health_manager, MetricIndicator

from .direct_loadbalancer import DirectLoadBalancer
from .discovery_service_loadbalancer import DiscoveryServiceLoadBalancer
from .local_service_loadabalancer import LocalServiceLoadBalancer
from .retry_loadbalancer import RetryLoadBalancer
from ...discovery import get_discovery

from .._rpc_types import RpcServiceInfo, EndPoint
from ._lb_types import LoadBalancer, ServiceLoadBalancer

# gateway = GatewayLoadBalancer()
_direct = DirectLoadBalancer()


_load_balance_cache: Dict[str, ServiceLoadBalancer] = {}
_local_service: Dict[str, EndPoint] = {}
_dependencies: Dict[str, bool] = {}


async def _metric_endpoint():
    return JSONResponse([v for v in _load_balance_cache.keys()])


get_health_manager().add_metric(MetricIndicator(
    # 相对于/actuator/metrics的路径，必须以'/'开头
    path='/dependencies',
    endpoint=_metric_endpoint
))

CHECK_ALIVE_INTERVAL = 5

# 定时检查目标服务的信息是否还有效，如果不存在了，则会移除


def print_time():
    invalidates = []
    for serviceId, lb in _load_balance_cache.items():
        if not lb.check_alive():
            invalidates.append(serviceId)

    # 存在失效的服务，从缓存中移除
    if len(invalidates) > 0:
        for service_id in invalidates:
            del _load_balance_cache[service_id]


# 指定1秒后执行print_time函数
t = Timer(CHECK_ALIVE_INTERVAL, print_time)
t.start()


app_name = 'unknown'
product_line = 'unknown'


def _load_config():
    configure = get_configure()
    config = configure.get_config('mctech.rpc.service')
    global app_name, product_line
    app_name = configure.get_app_info().get('name') or 'unknown'

    k8s = os.environ.get("KUBERNETES_SERVICE_HOST")
    if k8s:
        # 在k8s运行环境，服务名称使用k8s带名字空间的完整名称
        product_line = os.environ.get("PRODUCT_LINE") or 'unknown'

    for service_id, service_url in config.items():
        result = urlparse(service_url)
        _local_service[service_id] = EndPoint(
            id=service_id,
            vip_address=False,
            host=result.hostname,
            port=result.port
        )


def create(retry: int, service: RpcServiceInfo) -> LoadBalancer:
    if not app_name:
        _load_config()

    retry = retry or 0
    rpc_type = service.rpc_type if service.rpc_type else 'internal'

    # lb: LoadBalancer
    # if type == 'gateway':
    #   lb = gateway
    # el
    if rpc_type == 'internal':
        lb = _direct if not service.service_id else _get_service_loadbalancer(service.service_id)
    else:
        raise RuntimeError('不支持的调用方式')

    if retry > 0:
        lb = RetryLoadBalancer(lb)
    return lb


def _get_service_loadbalancer(service_id: str) -> ServiceLoadBalancer:
    lb = _load_balance_cache.get(service_id)
    if not lb:
        inst = _local_service.get(service_id)
        if inst:
            lb = LocalServiceLoadBalancer(inst, app_name, product_line)
        elif not get_discovery().isLocal:
            lb = DiscoveryServiceLoadBalancer(service_id, app_name, product_line)
        else:
            raise RuntimeError(
                "本地调用配置中未找到service_id为: '%s'的服务" % service_id
            )

        _load_balance_cache[service_id] = lb
        _dependencies[service_id] = True
    return lb


def get_service_info(service_id: str) -> EndPoint:
    '''
    获取内部服务注册信息
    '''
    lb = _get_service_loadbalancer(service_id)
    inst = lb.choose_one()
    return inst


__all__ = [
    "create",
    "get_service_info"
]



from __future__ import absolute_import
import math
import py_eureka_client.eureka_client as eureka_client

from time import sleep
from typing import List, Dict, Optional
from random import random
from log4py import logging

from .abstarct_rule import AbstractRule

log = logging.getLogger('python.cloud.rpc.rule')


class RoundRobinRule(AbstractRule):
    def check_alive(self):
        apps = self._eureka.applications.get_application(self._app_id_upper)
        return len(apps.up_instances) > 0

    def _choose(self):
        if not self.check_alive():
            return None

        app = self._eureka.applications.get_application(self._app_id_upper)
        index = 0 if len(app.up_instances) == 1 else math.floor(random() * len(app.up_instances))
        inst = app.up_instances[index]
        return inst


class WeightedResponseTimeRule(RoundRobinRule):
    _accumulated_weights: List[float]
    _stats: Dict[str, "ServerStats"]

    def __init__(self, app_id, eureka: eureka_client.EurekaClient):
        super().__init__(app_id, eureka)
        self._accumulated_weights = []
        self._stats = {}

    def _choose(self):
        # /** @type {import('@mctech/eureka-js-client').EurekaClient.EurekaInstanceConfig} */
        server: Optional[eureka_client.Instance] = None
        while not server:
            # get hold of the current reference in case it is changed from the other thread
            current_weights = self._accumulated_weights
            app = self._eureka.applications.get_application(self._app_id)

            server_count = len(app.up_instances)
            if server_count == 0:
                return None

            server_index = 0
            # last one in the list is the sum of all weights
            max_total_weight = 0 if len(current_weights) == 0 else current_weights[-1]
            # No server has been hit yet and total weight is not initialized
            # fallback to use round robin
            if max_total_weight < 0.001 or server_count != len(current_weights):
                server = super()._choose()
                if not server:
                    return server
            else:
                # generate a random weight between 0 (inclusive) to maxTotalWeight(exclusive)
                random_weight = random() * max_total_weight
                # pick the server index based on the randomIndex
                n = 0
                for d in current_weights:
                    if d >= random_weight:
                        server_index = n
                        break
                    else:
                        n = n + 1

                server = app.up_instances[server_index]

            if not server:
                # 短暂停顿下
                sleep(0.001)
        return server

    def compute(self):
        if log.isEnabledFor(logging.DEBUG):
            log.debug('Weight adjusting job started')
        total_response_time = 0
        app = self._eureka.applications.get_application(self._app_id)

        # find maximal 95 % response time
        for inst in app.up_instances:
            # this will automatically load the stats if not in cache
            ss = self._stats.get(inst.instanceId)
            if not ss:
                continue
            total_response_time += ss.get_response_time_avg()
        # weight for each server is (sum of responseTime of all servers - responseTime)
        # so that the longer the response time, the less the weight and the less likely to be chosen
        weight_so_far = 0.0

        # create new list and hot swap the reference
        final_weights: List[float] = []
        for inst in app.up_instances:
            ss = self._stats.get(inst.instanceId)
            if not ss:
                continue
            weight = total_response_time - ss.get_response_time_avg()
            weight_so_far += weight
            final_weights.append(weight_so_far)
        self._accumulated_weights = final_weights

    def record_stats(self, server_id: str, response_time: float):
        server_stats = self._stats.get(server_id)
        if not server_stats:
            server_stats = ServerStats(server_id)
            self._stats[server_id] = server_stats
        server_stats.note_response_time(response_time)


class ServerStats:
    id: str
    # 总请求数
    num_values: int
    # 总用时
    sum_values: float

    def __init__(self, id: str):
        self.id = id
        self.num_values = 0
        self.sum_values = 0

    def note_response_time(self, response_time: float):
        self.num_values = self.num_values + 1
        self.sum_values += response_time

    def get_response_time_avg(self):
        return self.sum_values / self.num_values



from __future__ import absolute_import
import websocket
from typing import Iterable, AsyncIterable

from ._lb_types import ServiceLoadBalancer
from ..request_context import InternalRequestContext
from ..ep import EndPointRequest

from .._rpc_types import Invoker, RpcServiceInfo, EndPoint, MixedMethod, RpcResponseResult


class LocalServiceLoadBalancer(ServiceLoadBalancer):
    '''
    显示指定了调用的目标地址的LoadBalancer实现方式
    '''

    def __init__(self, inst: EndPoint, app_name: str, product_line: str):
        super().__init__()
        self._inst = inst
        self.app_name = app_name
        self.product_line = product_line

    def check_alive(self) -> bool:
        return True

    def choose_one(self) -> EndPoint:
        return self._inst

    def create_context(self,
                       method: MixedMethod,
                       invoker: Invoker,
                       service: RpcServiceInfo) -> InternalRequestContext:
        return InternalRequestContext(method, invoker, service, self.app_name, self.product_line)

    async def execute_async(self, req: EndPointRequest) -> RpcResponseResult:
        result_data = await req.execute_async(self._inst)
        return result_data

    async def stream_async(self, req: EndPointRequest) -> AsyncIterable[bytes]:
        result_stream = await req.stream_async(self._inst)
        return result_stream

    async def ws_async(self, req: EndPointRequest) -> websocket.WebSocket:
        conn = await req.ws_async(self._inst)
        return conn

    def execute(self, req: EndPointRequest) -> RpcResponseResult:
        result_data = req.execute(self._inst)
        return result_data

    def stream(self, req: EndPointRequest) -> Iterable[bytes]:
        result_stream = req.stream(self._inst)
        return result_stream

    def ws(self, req: EndPointRequest) -> websocket.WebSocket:
        conn = req.ws(self._inst)
        return conn
